#include<stdio.h>
#include<getopt.h>
#include<limits.h>
#include<stdint.h>

#include "platform.h"
#include "xil_printf.h"
#include "xil_cache.h"


#include "non_iid/utils.h"
#include "non_iid/most_common.h"
#include "non_iid/collision_test.h"
#include "non_iid/lz78y_test.h"
#include "non_iid/multi_mmc_test.h"
#include "non_iid/lag_test.h"
#include "non_iid/multi_mcw_test.h"
#include "non_iid/compression_test.h"
#include "non_iid/markov_test.h"

#define SAMPLE_SIZE 1000000

volatile uint32_t * axi_address_base;

#include "communication.h"
#include "instruction.h"
#include "non_iid_tests.h"


int main(){
	uint32_t i, j, r;
	uint64_t INS[64];   // Instruction buffer for cryptoprocessor.
	uint64_t fpga_memory_data[1024]; // This is a memory buffer in SW. This buffer is used to read contents from the BRAM-memory of FPGA.


	unsigned int number_of_ones=0;
	int run_again = 1;
	double min_entropy;
	int COUNTER, num_of_rbits;


	uint8_t *data = (uint8_t*) malloc(SAMPLE_SIZE*sizeof(uint8_t));	// Buffer where 1M random bits. They will be used to perform NIST tests in ARM.


	init_platform();
	axi_address_base   = (uint32_t *) 0x40000000;		// For PYNQ-z2 board



	// Load instruction memory of Cryptoprocessor with the program
	// to generate 512 random words in the BRAM. The 512*64 bits are read into data[].
	// This program will be use 31 times to generate 1M random bits in total.

	///////////// In the following part a statistical test is performed in FPGA ////////////////


	// Now ask FPGA compute the statistical test.
	init_INS_stat(INS);
	send64(INS, 0, 64, 1); // send INStructions to FPGA using send64() function.
	delay(1000);
	exeIns();
	delay(1000);

	// During the execution of the program, it will print the 'error' flag.
	// The error flag is 1 when the min_entropy is found to be lower after the on-chip statistical test.





	///////////// In the following part a 1M random bits are generated in the FPGA and statistical tests are performed in ARM ////////////////

	init_INS_rng(INS);
	send64(INS, 0, 64, 1); 	// send INStructions to FPGA using send64() function.
	delay(1000);

	while(run_again){

		// The counter runs 31 times to generate 31*64*512 bits which is roughly 1M.
		// Each iteration of the COUNTER-loop generates 64*512 bits in the BRAM of FPGA.
		// The bits from fpga_memory_data buffer are stored into data[ ].
		// When 1M bits are available in data[], the NIST tests are called using non_iid_tests().

		num_of_rbits = 0;
		for(COUNTER=0; COUNTER<31; COUNTER++){

			// COUNTER=i: Generate 64*512 bits in BRAM and then read the random words.
			exeIns();
			delay(1000);
			receive64(fpga_memory_data, 0, 512);
			delay(1000);


			// Next, fill data[COUNTER*64*512 + ...] with the random bits from fpga_memory_data.
			uint64_t temp;
			uint64_t i_mod64;
			for(i=0; i<64*512; i++){
				temp = fpga_memory_data[i/64];
				i_mod64 = i%64;
				temp = temp>>i_mod64;
				temp = temp & 1;
				data[i+64*512*COUNTER] = (uint8_t) temp;
				number_of_ones = number_of_ones + (uint8_t) temp;
				if(i+64*512*COUNTER == SAMPLE_SIZE - 1) break;
			}
		}


		min_entropy = non_iid_tests(data);	// Performs NIST tests and returns the minimum entropy from all the tests.

		printf("Final min-entropy after all NIST tests in SW = %lf\n", min_entropy);


		// Just to see the random words, print them
		//for(i=0; i<512; i++)
		//	printf("fpga_memory_data[%d] = %llx\n", i, fpga_memory_data[i]);
		//printf("Number of ones in the TRNG string = %u\n", number_of_ones);

		printf("Run again [0/1]?\n");
		scanf("%d", &run_again);
	}


	free(data);



	Xil_DCacheFlush();
	Xil_DCacheDisable();

	printf("Finish\n");

	cleanup_platform();
	return 0;

}



